@@ -1,8 +1,8 @@
 /*
  * balance - a balancing tcp proxy
- * $Revision: 3.42 $
+ * $Revision: 3.52 $
  *
- * Copyright (c) 2000-2007,2008 by Thomas Obermair (obermair@acm.org)
+ * Copyright (c) 2000-2009,2010 by Thomas Obermair (obermair@acm.org)
  * and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany.
  * All rights reserved.
  *
@@ -13,6 +13,18 @@
  *
  * This program is dedicated to Richard Stevens...
  *
+ *  3.52
+ *    thanks to David J. Jilk from Standing Cloud, Inc. for the following:
+ *    added "nobuffer" functionality to interactive shell IO
+ *    added new "assign" interactive command
+ *    fixed locking bug
+ *  3.50
+ *    new option -6 forces IPv6 bind (hints.ai_family = AF_INET6)
+ *  3.49
+ *    ftok() patch applied (thanks to Vladan Djeric)  
+ *  3.48
+ *    Problems with setting IPV6_V6ONLY socket option are now handled 
+ *    more nicely with a syslog warning message 
  *  3.42
  *    Balance compiles now on systems where IPV6_V6ONLY is undefined
  *  3.35
@@ -26,7 +38,7 @@
  *  3.32
  *    /var/run/balance may already exist (thanks to Thomas Steudten)
  *  3.31
- *    TCP_NODELAY properly switched on (thank to Kurt J. Lidl). 
+ *    TCP_NODELAY properly switched on (thanks to Kurt J. Lidl). 
  *  3.30
  *    Code cleanups and fixes (thanks to Kurt J. Lidl)
  *  3.28
@@ -62,7 +74,8 @@
  *  3.10
  *    Bugfix for RedHat 7.2
  *  3.9
- *    Moved rendezvous file to /var/run and cleaned main(), thanks to Kayne Naughton
+ *    Moved rendezvous file to /var/run and cleaned main(), thanks to
+ *    Kayne Naughton
  *  3.8
  *    move to sigaction(), thanks to Kayne Naughton
  *  3.5
@@ -93,8 +106,8 @@
 
 #include <balance.h>
 
-const char *balance_rcsid = "$Id: balance.c,v 3.42 2008/04/08 17:37:09 tommy Exp $";
-static char *revision = "$Revision: 3.42 $";
+const char *balance_rcsid = "$Id: balance.c,v 3.52 2010/02/14 18:35:09 t Exp $";
+static char *revision = "$Revision: 3.52 $";
 
 static int release;
 static int subrelease;
@@ -120,6 +133,7 @@ static int foreground = 0;
 static int packetdump = 0;
 static int interactive = 0;
 static int shmmapfile = 0;
+static int bindipv6 = 0;
 
 static int sockbufsize = 32768;
 
@@ -134,11 +148,15 @@ static struct timeval save_tmout = { 0, 0 }; /* seconds, microseconds */
 int create_serversocket(char* node, char* service) {
   struct addrinfo hints;
   struct addrinfo *results;
-  int srv_socket, status, sockoptoff, sockopton;
+  int srv_socket, status, sockopton, sockoptoff;
 
   bzero(&hints, sizeof(hints));
   hints.ai_flags = AI_PASSIVE;
-  hints.ai_family = AF_UNSPEC;
+  if(bindipv6) {
+    hints.ai_family = AF_INET6;
+  } else {
+    hints.ai_family = AF_UNSPEC;
+  }
   hints.ai_socktype = SOCK_STREAM;
   hints.ai_protocol = IPPROTO_TCP;
 
@@ -166,8 +184,7 @@ int create_serversocket(char* node, char* service) {
 #if defined(IPV6_V6ONLY)
   status = setsockopt(srv_socket, IPPROTO_IPV6, IPV6_V6ONLY, (char*) &sockoptoff, sizeof(sockoptoff));
   if(status < 0) {
-    perror("setsockopt(IPV6_V6ONLY=0)");
-    exit(EX_OSERR);
+    syslog(LOG_WARNING,"setsockopt(IPV6_V6ONLY=0) failed");
   }
 #endif
 
@@ -326,10 +343,53 @@ void *shm_malloc(char *file, int size)
   }
 
   if(!shmmapfile){
+
+#if defined (__SVR4) && defined (__sun)
+
+    /* vdjeric:
+       Solaris ftok() causes frequent collisions because it uses
+       only the lower 12 bits of the inode number in the 'key'.
+       See: http://bugs.opensolaris.org/bugdatabase/view_bug.do?bug_id=4265917
+    */
+    
+    FILE *rendezvousfp = NULL;
+    struct timeval ct;
+    long int seed;
+    int i;
+
+    if ((rendezvousfp = fdopen(rendezvousfd, "w+")) == NULL) {
+      perror("fdopen");
+      exit(EX_OSERR);
+    }
+
+    if ((fscanf(rendezvousfp, "0x%x\n", &key)) <= 0) {
+      gettimeofday(&ct, NULL);
+      seed = ct.tv_usec * getpid(); 
+      srand(seed);
+
+      /* Solaris rand() returns values between 0 and 0x7fff, 
+         so generate key byte by byte */
+      key = 0;
+      for (i = 0; i < sizeof(key); i++) {
+          key = (key << 8) | (rand() & 0xff);
+      }
+
+      if(fseek(rendezvousfp, 0, SEEK_SET) == -1) {
+        perror("fseek");
+        exit(EX_OSERR);
+      }
+      if (fprintf(rendezvousfp, "0x%08x\n", key) == -1) {
+        perror("fprintf");
+        exit(EX_OSERR);
+      }
+      fflush(rendezvousfp);
+    }
+#else
     if ((key = ftok(file, 'x')) == -1) {
       perror("ftok");
       exit(EX_SOFTWARE);
     }
+#endif
 
     if ((shmid = shmget(key, size, 0644 | IPC_CREAT)) == -1) {
       perror("shmget");
@@ -901,7 +961,7 @@ void usage(void)
 
 
   fprintf(stderr, "  this is balance %d.%d\n", release, subrelease);
-  fprintf(stderr, "  Copyright (c) 2000-2007,2008\n");
+  fprintf(stderr, "  Copyright (c) 2000-2009,2010\n");
   fprintf(stderr, "  by Inlab Software GmbH, Gruenwald, Germany.\n");
   fprintf(stderr, "  All rights reserved.\n");
   fprintf(stderr, "\n");
@@ -955,7 +1015,8 @@ void background(void) {
 #else
   setpgrp();
 #endif
-  chdir("/");
+  if(chdir("/") <0) 
+    fprintf(stderr, "cannot chdir\n");
   close(0);
   close(1);
   close(2);
@@ -1080,6 +1141,12 @@ int shell(char *argument)
   char line[MAXINPUTLINE];
   char *command;
 
+  // DJJ, Standing Cloud, Inc.
+  //    In interactive mode, don't buffer stdout/stderr, so that
+  //    other programs can operate balance through I/O streams
+  setvbuf(stdout, NULL, _IONBF, 0);
+  setvbuf(stderr, NULL, _IONBF, 0);
+
   if (common->release == 0) {
     printf("no master process, exiting.\n");
     exit(EX_UNAVAILABLE);
@@ -1147,33 +1214,22 @@ int shell(char *argument)
       } else if (mycmp(command, "help") || mycmp(command, "?")) {
 	printf("available commands:\n");
 
-	printf
-	    ("  create <host> <port>  creates a channel in the current group\n");
-	printf
-	    ("  disable <channel>     disables specified channel in current group\n");
-	printf
-	    ("  enable <channel>      enables channel in current group\n");
-	printf
-	    ("  group <group>         changes current group to <group>\n");
-	printf
-	    ("  hash                  sets distribution scheme of current group to Hash\n");
-	printf("  help                  prints this message\n");
-	printf
-	    ("  kill                  kills master process and quits interactive mode\n");
-	printf
-	    ("  maxc <channel> <maxc> specifies new maxc for channel of current group\n");
-	printf
-	    ("  mrtg-bytes <grp> <ch> print bytes in/out in MRTG format\n");
-	printf
-	    ("  mrtg-conns <grp> <ch> print total connections in MRTG format\n");
-	printf("  quit                  quit interactive mode\n");
-	printf
-	    ("  reset <channel>       reset all counters of channel in current group\n");
-	printf
-	    ("  rr                    sets distribution scheme of current group to Round Robin\n");
-	printf
-	    ("  show                  show all channels in all groups\n");
-	printf("  version               show version id\n");
+	printf("  create <host> <port>           creates a channel in the current group\n");
+        printf("  assign <channel> <host> <port> reassigns a channel in the current group\n");
+	printf("  disable <channel>              disables specified channel in current group\n");
+	printf("  enable <channel>               enables channel in current group\n");
+	printf("  group <group>                  changes current group to <group>\n");
+	printf("  hash                           sets distribution scheme of current group to Hash\n");
+	printf("  help                           prints this message\n");
+	printf("  kill                           kills master process and quits interactive mode\n");
+	printf("  maxc <channel> <maxc>          specifies new maxc for channel of current group\n");
+	printf("  mrtg-bytes <grp> <ch>          print bytes in/out in MRTG format\n");
+	printf("  mrtg-conns <grp> <ch>          print total connections in MRTG format\n");
+	printf("  quit                           quit interactive mode\n");
+	printf("  reset <channel>                reset all counters of channel in current group\n");
+	printf("  rr                             sets distribution scheme of current group to Round Robin\n");
+	printf("  show                           show all channels in all groups\n");
+	printf("  version                        show version id\n");
 
       } else if (mycmp(command, "kill")) {
 	kill(common->pid, SIGKILL);
@@ -1194,14 +1250,14 @@ int shell(char *argument)
 	  if (n < 0 || n >= grp_nchannels(common, currentgroup)) {
 	    printf("no such channel %d\n", n);
 	  } else {
-	    c_writelock(0, n);
+	    c_writelock(currentgroup, n);
 	    if (chn_status(common, currentgroup, n) == 0) {
 	      printf("channel %d already disabled\n", n);
 	    } else {
 	      chn_status(common, currentgroup, n) = 0;
 	      printf("channel %d disabled\n", n);
 	    }
-	    c_unlock(0, n);
+	    c_unlock(currentgroup, n);
 	  }
 	} else {
 	  printf("syntax error\n");
@@ -1298,7 +1354,41 @@ int shell(char *argument)
 	}
 	b_unlock();
 
-      } else if (mycmp(command, "maxc")) {
+    } else if (mycmp(command, "assign")) {
+        char *arg1, *arg2, *arg3;
+
+          if ((arg1 = strtok(NULL, " \t\n")) != NULL) {
+        int chn = atoi(arg1);
+            if (chn < 0 || chn >= MAXCHANNELS
+                      || chn >= grp_nchannels(common, currentgroup)) {
+               printf("unknown channel\n");
+        } else {
+            c_writelock(currentgroup, chn);
+            if (chn_status(common, currentgroup, chn) != 0) {
+               printf("channel must be disabled to assign new address\n");
+            } else if ((arg2 = strtok(NULL, " \t\n")) != NULL) {
+                if ((arg3 = strtok(NULL, " \t\n")) != NULL) {
+                   if (setaddress_noexitonerror
+                         (&chn_ipaddr(common, currentgroup, chn),
+                          &chn_port(common, currentgroup, chn),
+                          arg2, getport(arg3))) {
+                       printf("channel reassigned\n");
+                   } else {
+                       printf("invalid address\n");
+                   }
+                } else {
+                   printf("syntax error\n");
+                }
+            } else {
+                printf("syntax error\n");
+            }
+            c_unlock(currentgroup, chn);
+        }
+      } else {
+        printf("syntax error\n");
+      }
+
+    } else if (mycmp(command, "maxc")) {
 	char *arg1, *arg2;
 	b_writelock();
 	if ((arg1 = strtok(NULL, " \t\n")) != NULL) {
@@ -1319,7 +1409,7 @@ int shell(char *argument)
 	}
 	b_unlock();
 
-      } else if (mycmp(command, "mrtg-bytes")) {
+    } else if (mycmp(command, "mrtg-bytes")) {
 	char *arg1, *arg2;
 	int mygroup, mychannel;
 	b_writelock();
@@ -1425,8 +1515,11 @@ int main(int argc, char *argv[])
   connect_timeout = DEFAULTTIMEOUT;
   initialize_release_variables();
 
-  while ((c = getopt(argc, argv, "c:b:B:t:T:adfpiHM")) != EOF) {
+  while ((c = getopt(argc, argv, "c:b:B:t:T:adfpiHM6")) != EOF) {
     switch (c) {
+    case '6':
+      bindipv6 = 1;
+      break;
     case 'a':
       autodisable = 1;
       break;
@@ -1542,9 +1635,9 @@ int main(int argc, char *argv[])
    */
 
   if (bindhost != NULL) {
-    snprintf(bindhost_address, FILENAMELEN, bindhost);
+    snprintf(bindhost_address, FILENAMELEN, "%s", bindhost);
   } else {
-    snprintf(bindhost_address, FILENAMELEN, "0.0.0.0");
+    snprintf(bindhost_address, FILENAMELEN, "%s", "0.0.0.0");
   }
 
   stat(SHMDIR, &buffer);
@@ -1596,6 +1689,8 @@ int main(int argc, char *argv[])
     shell(argument);
   }
 
+  openlog("Balance", LOG_ODELAY | LOG_PID | LOG_CONS, LOG_DAEMON);
+
   // Open a TCP socket (an Internet stream socket).
 
   sockfd = create_serversocket(bindhost, argv[0]);
@@ -1611,7 +1706,6 @@ int main(int argc, char *argv[])
     background();
   }
 
-  openlog("Balance", LOG_ODELAY | LOG_PID | LOG_CONS, LOG_DAEMON);
   common = makecommon(argc, argv, source_port);
 
   for (;;) {
